/***************************************************************************
 *   Copyright (C) 2006 by Ilya Kotov                                      *
 *   forkotov02@hotmail.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <QWidget>
#include <QFile>
#include <QDir>
#include <QtAlgorithms>
#include <QFileInfo>

#include <time.h>

#include <decoder.h>
#include <decoderfactory.h>

#include "fileloader.h"
#include "playlistmodel.h"
#include "mediafile.h"

PlayListModel::PlayListModel ( QObject *parent )
        : QObject ( parent ) , m_selection()
{
    qsrand(time(0));
    readSettings();
    m_current = 0;
    is_shuffle_list = false;
    m_shuffled_current = 0;
    is_repeatable_list = false;
}

PlayListModel::~PlayListModel()
{
    writeSettings();
    clear();
}

void PlayListModel::load ( MediaFile *file )
{
    if (m_files.isEmpty())
        m_currentItem = file;

    m_files << file;

    if (is_shuffle_list)
    {
        m_shuffled_indexes << m_files.count() - 1;
        m_shuffled_indexes.swap(m_shuffled_indexes.count() - 1,qrand()%m_files.size());
    }

    emit listChanged();
}

int PlayListModel::count()
{
    return m_files.size();
}

MediaFile* PlayListModel::currentItem()
{
    if ( m_files.isEmpty() )
        return 0;
    else
        return m_files.at ( m_current );
}

int PlayListModel::currentRow()
{
    return m_current;
    //return m_files.indexOf(m_currentItem);
}

bool PlayListModel::setCurrent ( int c )
{
    if ( c > count()-1 )
        return FALSE;
    m_current = c;
    m_currentItem = m_files.at(c);
    emit currentChanged();
    emit listChanged();
    return TRUE;
}


// we should hold counter and go through m_shuffled_indexes. The next value of
// m_shuffled_indexes[counter] becomes m_current. If counter reaches the end or begin of m_shuffled_indexes - do STOP.
// If new songs has been added/removed - update m_shuffled_indexes.
bool PlayListModel::next()
{
    if ( m_current == count()-1 || count() == 0)
    {
        if (is_repeatable_list && !count() == 0)
            m_current = -1;
        else if (!is_shuffle_list)
            return FALSE;
    }
    if (is_shuffle_list)
    {
        m_shuffled_current++;
        if (m_shuffled_current > m_shuffled_indexes.count() - 1)
        {
            if (!is_repeatable_list)
            {
                m_shuffled_current --;
                return FALSE;
            }
            else
                prepareForShufflePlaying(true);
        }

        m_current = m_shuffled_indexes.at(m_shuffled_current);
    }
    else
    {
        m_current++;
        m_currentItem = m_files.at(m_current);
    }
    emit currentChanged();
    emit listChanged();
    return TRUE;

}

bool PlayListModel::previous()
{
    if ( m_current < 1 && !is_shuffle_list)
        return FALSE;

    if (is_shuffle_list)
    {
        m_shuffled_current --;
        if (m_shuffled_current < 0)
        {
            if (!is_repeatable_list)
            {
                m_shuffled_current = 0;
                return FALSE;
            }
            else
                prepareForShufflePlaying(true);
        }

        m_current = m_shuffled_indexes.at(m_shuffled_current);
    }
    else
    {
        m_current--;
        m_currentItem = m_files.at(m_current);
    }
    emit currentChanged();
    emit listChanged();
    return TRUE;
}

void PlayListModel::clear()
{
    m_current = 0;
    while ( !m_files.isEmpty() )
        delete m_files.takeFirst();

    m_shuffled_indexes.clear();
    m_shuffled_current = 0;

    emit listChanged();
}

void PlayListModel::clearSelection()
{
    for ( int i = 0; i<m_files.size(); ++i )
        m_files.at ( i )->setSelected ( FALSE );
    emit listChanged();
}

QList <QString> PlayListModel::getTitles ( int b,int l )
{
    QList <QString> m_titles;
    for ( int i = b; ( i < b + l ) && ( i < m_files.size() ); ++i )
        m_titles << m_files.at ( i )->title();
    return m_titles;
}

QList <QString> PlayListModel::getTimes ( int b,int l )
{
    QList <QString> m_times;
    for ( int i = b; ( i < b + l ) && ( i < m_files.size() ); ++i )
        m_times << QString ( "%1" ).arg ( m_files.at ( i )->length() /60 ) +":"
        +QString ( "%1" ).arg ( m_files.at ( i )->length() %60/10 ) +
        QString ( "%1" ).arg ( m_files.at ( i )->length() %60%10 );
    return m_times;
}

bool PlayListModel::isSelected ( int row )
{
    return m_files.at ( row )->isSelected();
}

void PlayListModel::setSelected ( int row, bool yes )
{
    m_files.at ( row )->setSelected ( yes );
}

void PlayListModel::removeSelected()
{
    int i = 0;
    while ( !m_files.isEmpty() && i<m_files.size() )
    {
        if ( m_files.at ( i )->isSelected() )
        {
            delete m_files.takeAt ( i );

            if ( m_current >= i && m_current!=0 )
                m_current--;
        }
        else
            i++;
    }
    if (!m_files.isEmpty())
        m_currentItem = m_files.at(m_current);
    //prepareForShufflePlaying(true);

    if(is_shuffle_list)
        prepareForShufflePlaying(true);

    emit listChanged();
}

void PlayListModel::removeUnselected()
{
    int i = 0;
    while ( !m_files.isEmpty() &&i<m_files.size() )
    {
        if ( !m_files.at ( i )->isSelected() )
        {
            delete m_files.takeAt ( i );
            if ( m_current >= i && m_current!=0 )
                m_current--;
        }
        else
            i++;
    }

    if (!m_files.isEmpty())
        m_currentItem = m_files.at(m_current);

    if(is_shuffle_list)
        prepareForShufflePlaying(true);

    emit listChanged();
}

void PlayListModel::invertSelection()
{
    for ( int i = 0; i<m_files.size(); ++i )
        m_files.at ( i )->setSelected ( !m_files.at ( i )->isSelected() );
    emit listChanged();
}

void PlayListModel::selectAll()
{
    for ( int i = 0; i<m_files.size(); ++i )
        m_files.at ( i )->setSelected ( TRUE );
    emit listChanged();
}

void PlayListModel::showDetails()
{
    for ( int i = 0; i<m_files.size(); ++i )
    {
        if ( m_files.at ( i )->isSelected() )
        {
            DecoderFactory *fact = Decoder::findFactory ( m_files.at ( i )->path() );
            if ( fact )
                fact->showDetails ( 0, m_files.at ( i )->path() );

            return;
        }
    }

}

void PlayListModel::readSettings()
{
    QFile file ( QDir::homePath() +"/.qmmp/playlist.txt" );
    file.open ( QIODevice::ReadOnly );
    file.setTextModeEnabled ( TRUE );
    QByteArray line;
    m_files.clear();
    while ( !file.atEnd () )
    {
        line = file.readLine();
        load(new MediaFile ( QString::fromUtf8 ( line ).trimmed () ));
    }
    file.close ();
}

void PlayListModel::writeSettings()
{
    QFile file ( QDir::homePath() +"/.qmmp/playlist.txt" );
    file.open ( QIODevice::WriteOnly );
    foreach ( MediaFile* m, m_files )
    file.write ( m->path().toUtf8 () +"\n" );
    file.close ();
}

void PlayListModel::addFile(const QString& path)
{
    if (path.isEmpty ())
        return;
    if (Decoder::supports(path))
        load(new MediaFile(path));
}

void PlayListModel::addFiles(const QStringList &files)
{
    if (files.isEmpty ())
        return;
    foreach(QString s, files)
    {
        if (Decoder::supports(s))
            load(new MediaFile(s));
    }
}

void PlayListModel::addDirectory(const QString& s)
{
    FileLoader fl(this);
    QFileInfoList list = fl.getFileList(s);

    for (int i = 0; i < list.size(); ++i)
    {
        QFileInfo fileInfo = list.at(i);
        if (Decoder::supports(fileInfo.absoluteFilePath ()))
            load(new MediaFile(fileInfo.absoluteFilePath ()));

    }
}

int PlayListModel::firstSelectedUpper(int row)
{
    for (int i = row - 1;i >= 0;i--)
    {
        if (isSelected(i))
            return i;
    }
    return -1;
}

int PlayListModel::firstSelectedLower(int row)
{
    for (int i = row + 1;i < count() ;i++)
    {
        if (isSelected(i))
            return i;
    }
    return -1;
}

void PlayListModel::moveItems( int from, int to )
{
    if (from == to)
        return;

    SimpleSelection sel = getSelestion( from );

    if (sel.isValid())
    {
        int top_diff = sel.m_anchor - sel.m_top;

        QList<MediaFile*>temporary_storage = m_files.mid(sel.m_top , sel.count());

        for (int i = sel.m_top;i <= sel.m_bottom;i++)
            m_files.removeAt(sel.m_top);

        for (int i = 0; i < temporary_storage.count();i++)
            m_files.insert(to + i - top_diff,temporary_storage[i]);

        m_current = m_files.indexOf(m_currentItem);

        emit listChanged();
    }
}



int PlayListModel::topmostInSelection( int row)
{
    if ( row == 0)
        return 0;

    for (int i = row - 1;i >= 0;i--)
    {
        if (isSelected(i))
            continue;
        else
            return i + 1;
    }
    return 0;
}

int PlayListModel::bottommostInSelection( int row )
{
    if (row >= m_files.count() - 1)
        return row;

    for (int i = row + 1;i < count() ;i++)
    {
        if (isSelected(i))
            continue;
        else
            return i - 1;
    }
    return count() - 1;
}

const SimpleSelection& PlayListModel::getSelestion(int row )
{
    m_selection.m_top = topmostInSelection( row );
    m_selection.m_anchor = row;
    m_selection.m_bottom = bottommostInSelection( row );
    return m_selection;
}

QList<int> PlayListModel::getSelectedRows() const
{
    QList<int>selected_rows;
    for (int i = 0;i<m_files.count();i++)
    {
        if (m_files[i]->isSelected())
        {
            selected_rows.append(i);
        }
    }
    return selected_rows;
}

QList< MediaFile * > PlayListModel::getSelectedItems() const
{
    QList<MediaFile*>selected_items;
    for (int i = 0;i<m_files.count();i++)
    {
        if (m_files[i]->isSelected())
        {
            selected_items.append(m_files[i]);
        }
    }
    return selected_items;
}

void PlayListModel::addToQueue()
{
    QList<MediaFile*> selected_items = getSelectedItems();
    foreach(MediaFile* file,selected_items)
    {
        if (isQueued(file))
            m_queued_songs.removeAt(m_queued_songs.indexOf(file));
        else
            m_queued_songs.append(file);
    }
    emit listChanged();
}

bool PlayListModel::isQueued(MediaFile* f) const
{
    return m_queued_songs.contains(f);
}

void PlayListModel::setCurrentToQueued()
{
    setCurrent(row(m_queued_songs.at(0)));
    m_queued_songs.pop_front();
}

bool PlayListModel::isEmptyQueue() const
{
    return m_queued_songs.isEmpty();
}

void PlayListModel::randomizeList()
{
    for (int i = 0;i < m_files.size();i++)
        m_files.swap(qrand()%m_files.size(),qrand()%m_files.size());

    m_current = m_files.indexOf(m_currentItem);
    emit listChanged();
}

void PlayListModel::reverseList()
{
    for (int i = 0;i < m_files.size()/2;i++)
        m_files.swap(i,m_files.size() - i - 1);

    m_current = m_files.indexOf(m_currentItem);
    emit listChanged();
}

////===============THE BEGINNING OF SORT IMPLEMENTATION =======================////

// First we'll implement bundle of static compare procedures
// to sort items in different ways
static bool _titleLessComparator(MediaFile* s1,MediaFile* s2)
{
    return s1->title() < s2->title();
}

static bool _titleGreaterComparator(MediaFile* s1,MediaFile* s2)
{
    return s1->title() > s2->title();
}

static bool _pathAndFilenameLessComparator(MediaFile* s1,MediaFile* s2)
{
    return s1->path() < s2->path();
}

static bool _pathAndFilenameGreaterComparator(MediaFile* s1,MediaFile* s2)
{
    return s1->path() > s2->path();
}

static bool _filenameLessComparator(MediaFile* s1,MediaFile* s2)
{
    QFileInfo i_s1(s1->path());
    QFileInfo i_s2(s2->path());
    return i_s1.baseName() < i_s2.baseName();
}

static bool _filenameGreaterComparator(MediaFile* s1,MediaFile* s2)
{
    QFileInfo i_s1(s1->path());
    QFileInfo i_s2(s2->path());
    return i_s1.baseName() > i_s2.baseName();
}

static bool _dateLessComparator(MediaFile* s1,MediaFile* s2)
{
    return s1->year() < s2->year();
}

static bool _dateGreaterComparator(MediaFile* s1,MediaFile* s2)
{
    return s1->year() > s2->year();
}

// This is main sort method
void PlayListModel::doSort(int sort_mode,QList<MediaFile*>& list_to_sort)
{
    QList<MediaFile*>::iterator begin;
    QList<MediaFile*>::iterator end;

    begin = list_to_sort.begin();
    end = list_to_sort.end();

    bool (*compareLessFunc)(MediaFile*,MediaFile*) = 0;
    bool (*compareGreaterFunc)(MediaFile*,MediaFile*) = 0;

    switch (sort_mode)
    {
    case TITLE:
        compareLessFunc = _titleLessComparator;
        compareGreaterFunc = _titleGreaterComparator;
        break;
    case FILENAME:
        compareLessFunc = _filenameLessComparator;
        compareGreaterFunc = _filenameGreaterComparator;
        break;
    case PATH_AND_FILENAME:
        compareLessFunc = _pathAndFilenameLessComparator;
        compareGreaterFunc = _pathAndFilenameGreaterComparator;
        break;
    case DATE:
        compareLessFunc = _dateLessComparator;
        compareGreaterFunc = _dateGreaterComparator;
        break;
        //qWarning("TODO Sort by Date: %s\t%d",__FILE__,__LINE__);
    default:
        compareLessFunc = _titleLessComparator;
        compareGreaterFunc = _titleGreaterComparator;
    }

    static bool sorted_asc = false;
    if (!sorted_asc)
    {
        qSort(begin,end,compareLessFunc);
        sorted_asc = true;
    }
    else
    {
        qSort(begin,end,compareGreaterFunc);
        sorted_asc = false;
    }

    m_current = m_files.indexOf(m_currentItem);
}

void PlayListModel::sortSelection(int mode)
{
    QList<MediaFile*>selected_items = getSelectedItems();
    QList<int>selected_rows = getSelectedRows();

    doSort(mode,selected_items);

    for (int i = 0;i < selected_rows.count();i++)
        m_files.replace(selected_rows[i],selected_items[i]);

    m_current = m_files.indexOf(m_currentItem);
    emit listChanged();
}

void PlayListModel::sort(int mode)
{
    doSort(mode,m_files);
    emit listChanged();
}

////=============== THE END OF SORT IMPLEMENTATION =======================////

void PlayListModel::prepareForShufflePlaying(bool val)
{
    is_shuffle_list = val;
    if (val)
    {
        m_shuffled_indexes.clear();
        m_shuffled_current = 0;
        for (int i = 0;i < m_files.count();i++)
            m_shuffled_indexes << i;
        for (int i = 0;i < m_files.count();i++)
            m_shuffled_indexes.swap(qrand()%m_files.size(),qrand()%m_files.size());

        if (m_shuffled_indexes.count() > 1 && (m_current == m_shuffled_indexes[0] || m_current == -1))
        {
            int tmp = m_shuffled_indexes.takeAt(0);
            m_shuffled_indexes.append(tmp);
        }
    }
    else
    {
        m_shuffled_indexes.clear();
        m_shuffled_current = 0;
    }
    //qWarning("TODO: %s\t%d",__FILE__,__LINE__);
}

void PlayListModel::prepareForRepeatablePlaying(bool val)
{
    is_repeatable_list = val;
}

